import asyncio
from typing import List, Any, Coroutine, Awaitable

from aflowey import AsyncFlow
from aflowey.single_executor import FlowOrListFlow, SingleFlowExecutor


class AsyncFlowExecutor:
    """
    Execute several flows in parallel

    >>>await (executor(flows) | flow).run()

    """

    def __init__(
        self, flows: List[AsyncFlow] = [], run_in_thread_pool: bool = False
    ) -> None:
        self.flows = []
        self.flows.append(flows)
        # to be run in parallel
        self.run_in_thread_pool = run_in_thread_pool

    def __or__(self, flow: List[AsyncFlow]) -> "AsyncFlowExecutor":
        """add a flow to execute in parallel"""
        self.flows.append(flow)

        return self

    @staticmethod
    async def _execute_one_flow(flow: AsyncFlow) -> Any:
        return await SingleFlowExecutor(flow).execute_flow()

    def exec_or_gather(self, flow: FlowOrListFlow) -> Awaitable[Any]:
        execute = self._execute_one_flow
        if isinstance(flow, list):
            return asyncio.gather(*[execute(flow) for flow in flow])
        return execute(flow)

    async def run(self, **kwargs: Any) -> Any:
        """main function to run stuff in parallel"""
        # if other flow run in parallel
        flows = [self.exec_or_gather(flow) for flow in self.flows]
        results = await asyncio.gather(*flows, **kwargs)
        return results

    @staticmethod
    def executor(flows: List[AsyncFlow]) -> "AsyncFlowExecutor":
        """create a new executor from one flow or array of flows"""
        return AsyncFlowExecutor(flows)


async_exec = aexec = AsyncFlowExecutor.executor



from aflowey.f import (
    F,
)
from aflowey.functions import (
    async_wrap,
    ensure_callable,
    side_effect,
    may_fail,
    spread,
    spread_args,
    spread_kw,
    spread_kwargs,
    f0,
    F0,
    f1,
    F1,
    erratic,
    breaker,
    lift,
    imp,
    impure,
    apartial,
    partial,
    identity,
)
from aflowey.async_flow import (
    AsyncFlow,
    aflow,
    async_flow,
    flog,
)
from aflowey.executor import (
    AsyncFlowExecutor,
    async_exec,
    aexec,
)

from aflowey.single_executor import (
    CANCEL_FLOW,
)

__all__ = [
    "F",
    "async_wrap",
    "ensure_callable",
    "side_effect",
    "f0",
    "F0",
    "f1",
    "F1",
    "erratic",
    "breaker",
    "may_fail",
    "lift",
    "imp",
    "impure",
    "AsyncFlow",
    "aflow",
    "async_flow",
    "AsyncFlowExecutor",
    "async_exec",
    "aexec",
    "apartial",
    "partial",
    "spread",
    "spread_args",
    "spread_kw",
    "spread_kwargs",
    "identity",
    "CANCEL_FLOW",
    "flog",
]



from copy import copy
from inspect import iscoroutinefunction
from typing import Any, Union, List, Dict

from loguru import logger

from aflowey import ensure_callable, F, async_wrap
from aflowey.functions import is_side_effect, get_name
from aflowey.types import Function


class AsyncFlow:
    """
    Describe an async flow chaining function

    >>>flow = (AsyncFlow() >> gen1 >> gen2 >> gen3)
    >>>await flow.run()
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # first arguments input
        self.args = args
        self.kwargs = kwargs

        # function to be executed in the flow
        self.aws: List[F] = []

        # intermediary store results
        self.steps: Dict[str, Any] = {}

    @staticmethod
    def ensure_coroutine_func(func: Union[Function, F]) -> F:
        side_effect_func = is_side_effect(func)
        named_func = get_name(func)

        func = ensure_callable(func)

        # ensuring F function are coroutine
        if not iscoroutinefunction(func):
            func = F(func) >> async_wrap

        # ensuring we have only F function
        if not isinstance(func, F):
            func = F(func)

        # resetting dunder custom attribute
        if side_effect_func:
            func.__side_effect__ = True  # type: ignore
        if named_func:
            func.__named__ = named_func  # type: ignore

        return func

    def __rshift__(
        self, aws: Union[List[Union[F, Function]], Union[F, Function]]
    ) -> "AsyncFlow":
        """add a new step to the flow

        Args:
            aws: list of callable or callable
        """
        if isinstance(aws, list):
            self.aws += [self.ensure_coroutine_func(a) for a in aws]
        else:
            self.aws.append(self.ensure_coroutine_func(aws))

        return self

    def __copy__(self) -> "AsyncFlow":
        """make a shallow copy"""
        aws = self.aws[:]
        args = self.args[:]
        kwargs = self.kwargs.copy()
        new_flow = self.__class__(*args, **kwargs)
        new_flow.aws = aws
        return new_flow

    async def run(self) -> Any:
        """run the flow

        Returns:
            coroutine
        """
        from aflowey.single_executor import SingleFlowExecutor

        return await SingleFlowExecutor(self).execute_flow()

    @staticmethod
    def log(log_str: str, print_arg: bool = False) -> Any:
        """utility function to log between steps, printing argument if needed"""

        async def wrapped(last_result: Any) -> Any:
            logger.info(log_str)
            if print_arg:  # pragma: no cover
                logger.info(last_result)
            return last_result

        return wrapped

    @staticmethod
    def from_args(*args: Any, **kwargs: Any) -> "AsyncFlow":
        """create a flow with given arguments as first input"""
        return AsyncFlow(*args, **kwargs)

    @staticmethod
    def empty() -> "AsyncFlow":
        """create an empty flow"""
        return AsyncFlow()

    @staticmethod
    def from_flow(flow: "AsyncFlow") -> "AsyncFlow":
        """create a new flow from given flow, copying it
        (args, kwargs, and aws functions)
        """
        assert isinstance(flow, AsyncFlow)
        return copy(flow)


# aliases
aflow = async_flow = AsyncFlow
flog = aflow.log



import functools
import inspect
from typing import Any, Callable, List, Union, Iterable, Optional, cast

from aflowey import F
from aflowey.types import Function


def async_wrap(func: F) -> F:
    """wrap the given into a coroutine function and try
    calling it
    """

    @functools.wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        async def _exec(function: Union[F, Function], *a: Any, **kw: Any) -> Any:
            current_result = function(*a, **kw)
            if inspect.iscoroutine(current_result):
                current_result = await current_result
                if isinstance(current_result, F):
                    return await _exec(current_result.func)
            return current_result

        return await _exec(func, *args, **kwargs)

    return F(wrapped)


def lift(func: Function, *args: Any, **kwargs: Any) -> F:
    """make a partial function of the given func and ensure
    it will work in an async context
    """
    new_func = cast(Callable[[], Any], functools.partial(func, *args, **kwargs))
    return F(new_func) >> async_wrap


apartial = partial = lift


def f1(func: Function, extractor: Optional[Function] = None) -> F:
    """wraps a one argument function (with arity 1) and allows
    to add an extractor to work on the input argument.
    """

    @functools.wraps(func)
    def wrapped(arg1: Any) -> Any:
        value = arg1 if extractor is None else extractor(arg1)
        return func(value)

    wrapped.__F1__ = True  # type: ignore
    return F(wrapped)


F1 = f1


def f0(func: Function) -> F:
    """create a new function from a 0 arity function (takes 0 args). The new
    function takes exactly one argument and does not pass it to the wrapped
    function. It allows using a 0 arity function in a flow relatively easily.
    """

    @functools.wraps(func)
    def wrapped(arg1: Any) -> Any:
        return func()

    wrapped.__F0__ = True  # type: ignore
    return F(wrapped)


F0 = f0


def may_fail(func: Function) -> Function:
    """ simply for readability"""
    return func


breaker = erratic = may_fail


def spread_args(func: Function) -> F:
    """create a function which takes an iterable of args
    and spread it into the given function"""

    @functools.wraps(func)
    def wrapped(args: Iterable[Any]) -> Any:
        # if to much args, slice the args to given length
        return func(*args)

    return F(wrapped)


spread = spread_args


def spread_kwargs(func: Function) -> F:
    """create a function which takes a mapping of kwargs
    and spread it into the given function"""

    @functools.wraps(func)
    def wrapped(**kwargs: Any) -> Any:
        arg_spec = inspect.getfullargspec(func)
        args = set(arg_spec.args + arg_spec.kwonlyargs) - set(["self"])
        new_kwargs = {key: kwargs[key] for key in args}
        return func(**new_kwargs)

    return F(wrapped)


spread_kw = spread_kwargs


def ensure_f(func: Function) -> F:
    if not isinstance(func, F):
        return F(func)
    return func


def make_impure(func: Union[Function, F]) -> F:
    # automatically create new function when the function
    # is a bound method and has no other input args
    func = ensure_f(func)
    func.__side_effect__ = True  # type: ignore
    return func


def side_effect(*func: Function) -> Union[List[F], F]:
    """
    take an array of function and tag it as side effects
    function
    """
    if len(func) == 1:
        return make_impure(func[0])
    return [make_impure(fu) for fu in func]


imp = impure = side_effect


def ensure_callable(x: Union[Any, Function]) -> Function:
    """ ensure a given args is a callable by returning a new callable if not"""
    if not callable(x):

        def wrapped(*args: Any, **kwargs: Any) -> Any:
            return x

        return cast(Function, wrapped)
    return cast(Function, x)


def named(func: Union[Function, F], name: str) -> F:
    if not callable(func):
        func = ensure_callable(func)
    func = ensure_f(func)
    func.__named__ = name  # type: ignore
    return func


def identity(x: Any) -> Any:
    return x


# helper functions
def is_f0(func: Function) -> bool:
    return hasattr(func, "__F0__")


def is_f1(func: Function) -> bool:
    return hasattr(func, "__F1__")


def is_side_effect(func: Function) -> bool:
    return hasattr(func, "__side_effect__")


def get_name(func: Function) -> bool:
    if hasattr(func, "__named__"):
        return func.__named__
    return ""



import functools
from typing import Any
from typing import Callable

from aflowey.types import Function


class F:
    """tiny wrapper around a function

    Args:
        func: callable

    """

    def __init__(self, func: Function) -> None:
        self.func = func
        functools.update_wrapper(self, func)

    def __rshift__(self, other: Callable[[Function], "F"]) -> "F":
        """equivalent of decoration"""
        return F(other(self.func))

    def __gt__(self, other: Function) -> "F":
        """function composition"""

        @functools.wraps(other)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            return other(self.func(*args, **kwargs))

        return F(wrapped)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """ simply call the inner function"""
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<F instance: {repr(self.func)}>"



from typing import Any, Union, List, cast

from loguru import logger

from aflowey import AsyncFlow, F
from aflowey.functions import is_f0, is_side_effect


class SingleFlowExecutor:
    """Single flow executor"""

    CANCEL_FLOW = object()

    def __init__(self, flow: AsyncFlow) -> None:
        self.flow = flow

    @staticmethod
    async def check_and_execute_flow_if_needed(
        maybe_flow: Union[Any, AsyncFlow]
    ) -> Any:
        """check if we have an async flow and execute it"""
        if isinstance(maybe_flow, AsyncFlow):
            return await SingleFlowExecutor(maybe_flow).execute_flow()
        return maybe_flow

    @staticmethod
    def need_to_cancel_flow(result: Any) -> bool:
        """check if we need to cancel flow checking sentinel"""
        if result is SingleFlowExecutor.CANCEL_FLOW:
            logger.debug("Received sentinel object, canceling flow...")
            return True
        return False

    @staticmethod
    def get_step_name(func: F, index: int) -> str:
        """get the step name"""
        if hasattr(func, "__named__"):
            return cast(str, func.__named__)  # type: ignore
        return str(index)

    def save_step(self, task: F, index: int, current_args: Any) -> None:
        """save step state in flow attribute"""
        self.flow.steps[self.get_step_name(task, index)] = current_args

    def _check_current_args_if_side_effect(self, first_aws: F, res: Any) -> Any:
        if is_side_effect(first_aws):
            return self.flow.kwargs if self.flow.kwargs else self.flow.args
        return res

    async def execute_first_step(self, first_aws: F) -> Any:
        """executing the first step"""
        if not self.flow.args and not self.flow.kwargs and is_f0(first_aws):
            self.flow.args = (None,)

        res = await first_aws(*self.flow.args, **self.flow.kwargs)
        current_args = self._check_current_args_if_side_effect(first_aws, res)
        # if flow run it
        current_args = await self.check_and_execute_flow_if_needed(current_args)

        # memorize name
        self.save_step(first_aws, 0, current_args)
        return current_args

    async def execute_flow(self) -> Any:
        """Main function to execute a flow"""
        if not self.flow.aws:
            return None

        # get first step
        first_aws = self.flow.aws[0]
        current_args = await self.execute_first_step(first_aws)

        # iterate over its tasks
        for index, task in enumerate(self.flow.aws[1:]):
            # get the result of a task
            result = await task(current_args)

            if is_side_effect(task):
                # side effect task, does not return a value
                self.save_step(task, index + 1, current_args)
                if self.need_to_cancel_flow(result):
                    break  # pragma: no cover
                continue  # pragma: no cover

            result = await self.check_and_execute_flow_if_needed(result)
            # cancel ?
            if self.need_to_cancel_flow(result):
                break

            current_args = result
            self.save_step(task, index + 1, current_args)
        # return current args that are the actual results
        return current_args


CANCEL_FLOW = SingleFlowExecutor.CANCEL_FLOW
FlowOrListFlow = Union[List[AsyncFlow], AsyncFlow]
